import java.util.ArrayList;
import java.util.Arrays;

/**
 * An EightsPuzzleWorldState object represents a possible state for
 * 8-puzzle-like problems. The problem consists of a rectangular grid of
 * numbered tiles with one gap, so the state is specified by a rectangular array
 * of integers.
 * 
 * @author John MacCormick, Dickinson College
 * @author Some code copied from Prof. Grant Braught; used with permission.
 * @version September 2010
 */
public class EightsPuzzleWorldState extends WorldState {

	/**
	 * A fixed value used to represent the presence of the gap, or "hole", in
	 * the puzzle board.
	 */
	public static final int HOLE_VALUE = 0;

	// A two-dimensional array representing the numbers on the tiles in the
	// puzzle.
	private int[][] board;

	// The row and column of the current location of the hole. Thus, we must
	// maintain the invariant that board[holeRow][holeColumn] == HOLE_VALUE.
	private int holeRow;
	private int holeColumn;

	/**
	 * Create a new puzzle state representing the given puzzle board.
	 * 
	 * @param board
	 *            An array representing the numbers on the tiles in the puzzle
	 *            board.
	 */
	public EightsPuzzleWorldState(int[][] board) {
		this.board = board;
		// initialize the holeRow, holeColumn fields correctly
		computeHoleLocation();
	}

	// Set the values of the holeRow, holeColumn fields to maintain the
	// invariant that board[holeRow][holeColumn] == HOLE_VALUE.
	private void computeHoleLocation() {
		boolean foundHole = false;
		for (int i = 0; i < board.length; i++) {
			for (int j = 0; j < board[0].length; j++) {
				if (board[i][j] == HOLE_VALUE) {
					holeRow = i;
					holeColumn = j;
					foundHole = true;
				}
			}
		}
		if (!foundHole) {
			throw new RuntimeException("No hole found in the puzzle.");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see WorldState#hashCode()
	 */
	@Override
	public int hashCode() {
		// Arrays.deepHashCode() computes a hash of all elements in a 2-D array,
		// which is what we need here.
		return Arrays.deepHashCode(board);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see WorldState#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		EightsPuzzleWorldState other = (EightsPuzzleWorldState) obj;
		// Arrays.deepEquals() tests for the equality of all elements in a 2-D
		// array, which is what we need here.
		return Arrays.deepEquals(board, other.board);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see WorldState#getValidActions()
	 */
	@Override
	public ArrayList<Action> getValidActions() {
		ArrayList<Action> actions = new ArrayList<Action>();

		// Can we slide up? If so, add the relevant action.
		if (holeRow == 0 || holeRow == 1) {
			actions.add(new EightsPuzzleAction(EightsPuzzleAction.Direction.Up));
		}

		// TODO: fill in the missing code
		
		return actions;
	}

	/**
	 * Clones the puzzle board.
	 * 
	 * @return A new 2-D array containing a copy of the calling object's puzzle
	 *         board.
	 */
	public int[][] cloneBoard() {
		int[][] tmp = board.clone();
		for (int row = 0; row < board.length; row++) {
			tmp[row] = board[row].clone();
		}
		return tmp;
	}

	/* (non-Javadoc)
	 * @see WorldState#apply(Action)
	 */
	@Override
	public WorldState apply(Action action) {
		EightsPuzzleAction theAction = (EightsPuzzleAction) action;
		int[][] newBoard = cloneBoard();
		int tileToMove = -1;
		int newHoleRow = holeRow;
		int newHoleColumn = holeColumn;

		switch (theAction.getDirection()) {
		case Up:
			newHoleRow++;
			break;

	    // TODO: fill in the missing code
			
		}

		tileToMove = board[newHoleRow][newHoleColumn];
		newBoard[holeRow][holeColumn] = tileToMove;
		newBoard[newHoleRow][newHoleColumn] = HOLE_VALUE;

		return new EightsPuzzleWorldState(newBoard);
	}

	/* (non-Javadoc)
	 * @see WorldState#toString()
	 */
	@Override
	public String toString() {
		StringBuilder representation = new StringBuilder();
		for (int row = 0; row < board.length; row++) {
			for (int col = 0; col < board[0].length; col++) {
				if (board[row][col] == HOLE_VALUE) {
					representation.append("   ");
				} else {
					representation.append(" " + board[row][col] + " ");
				}
			}
			representation.append("\n");
		}
		return representation.toString();
	}

	/**
	 * Return a reference to the calling object's puzzle board.
	 * @return A reference to the calling object's puzzle board.
	 */
	public int[][] getBoard() {
		return board;
	}

}
